Add LoadPath "TLC".

Require Import TLC.LibTactics.
Require Import Coq.Program.Equality.
Require Import Coq.Lists.List.
Require Import TLC.LibList.


Set Implicit Arguments.
Unset Bracketing Last Introduction Pattern.

Inductive exp :=
| e_var : nat -> exp
| e_lam : exp -> exp
| e_app : exp -> exp -> exp
| e_par : exp -> exp -> exp
| e_fst : exp -> exp
| e_snd : exp -> exp.

Notation "'#' n" := (e_var n) (at level 20) : exp_scope.
Notation "'λ' e" := (e_lam e) (at level 55) : exp_scope.
Notation "e₁ · e₂" := (e_app e₁ e₂) (at level 40, left associativity) : exp_scope.
Notation "e₁ ∥ e₂" := (e_par e₁ e₂) (at level 50, left associativity) : exp_scope.
Notation "'π₁' e" := (e_fst e) (at level 20) : exp_scope.
Notation "'π₂' e" := (e_snd e) (at level 20) : exp_scope.
Delimit Scope exp_scope with exp.

Inductive val :=
| v_clos : exp -> list val -> val
| v_pair : val -> val -> val.

Notation "⟨ v₁ , v₁' ⟩" := (v_pair v₁ v₁') : val_scope.
Notation "e \ σ " := (v_clos e σ) (at level 10) : val_scope.
Delimit Scope val_scope with val.
Arguments v_clos (e)%exp (σ)%val.

Definition sub := list val.

Inductive frame :=
| f_appL  : exp -> sub -> frame
| f_appR  : val -> frame
| f_pairL : exp -> sub -> frame
| f_pairR : val -> frame
| f_fst   : frame
| f_snd   : frame.

Definition stack := list frame.

Inductive head :=
| h_clo : exp -> sub -> head
| h_val : val -> head.

(* Check out bind scope *)
Notation "'V' v" := (h_val v) (at level 20) : hd_scope.
Notation "'C' e ∣ σ" := (h_clo e σ) (at level 20, σ at level 60) : hd_scope.
Arguments h_clo (e)%exp (σ)%val.
Arguments h_val (v)%val.
Delimit Scope head_scope with hd.

Definition conf := (head * stack)%type.

Inductive spdag :=
| sp0 : spdag
| sp1 : spdag
| spPar : spdag -> spdag -> spdag
| spSeq : spdag -> spdag -> spdag.

Notation "0" := sp0 : sp_scope.
Notation "1" := sp1 : sp_scope.
Notation "g₁ ∥ g₂" := (spPar g₁ g₂) (at level 50, left associativity) : sp_scope.
Notation "g₁ · g₂" := (spSeq g₁ g₂) (at level 40, left associativity) : sp_scope.
Delimit Scope sp_scope with sp.

Parameter τ : nat.

Section HelperDefs.
  Local Open Scope sp_scope.

  Fixpoint work s :=
    match s with
    | 0 => 0%nat
    | 1 => 1%nat
    | g₁ ∥ g₂ => 1 + τ + work g₁ + work g₂
    | g₁ · g₂ => work g₁ + work g₂
    end.

  Fixpoint span s :=
    match s with
    | 0 => 0%nat
    | 1 => 1%nat
    | g₁ ∥ g₂ => 1 + τ + max (span g₁) (span g₂)
    | g₁ · g₂ => span g₁ + span g₂
    end.

  Local Open Scope hd_scope.
  Local Open Scope exp_scope.

  Fixpoint hasPar k :=
    match k with
    | nil => false
    | f_pairL _ _ :: k => true
    | _ :: k => hasPar k
    end.

  Definition isPar (c : conf) :=
    match c with
    | (C (_ ∥ _) ∣ _, _) => true
    | _ => false
    end.

  Definition isParL (c : conf) :=
    match c with
    | (V _, f_pairL _ _ :: _) => true
    | _ => false
    end.

End HelperDefs.

Lemma hasPar_app_par k₁ k₂ e σ :
  hasPar (k₁ ++ f_pairL e σ :: k₂) = true.
Proof.
  induction k₁; simpl; [reflexivity | destruct a; tauto].
Qed.

Section Semantics.
  Local Open Scope hd_scope.
  Local Open Scope exp_scope.
  Local Open Scope val_scope.

  Reserved Notation "c₁ → c₂" (at level 70, no associativity).

  Inductive sstep : conf -> conf -> Prop :=
  | ss_var n σ v k (HS : nth_error σ n = Some v) :
      (C # n ∣ σ, k) → (V v, k)
  | ss_lam e σ k :
      (C λ e ∣ σ, k) → (V e \ σ, k)
  | ss_appL e₁ e₂ σ k :
      (C e₁ · e₂ ∣ σ, k) → (C e₁ ∣ σ, f_appL e₂ σ :: k)
  | ss_appR v e σ k :
      (V v, f_appL e σ :: k) → (C e ∣ σ, f_appR v :: k)
  | ss_app e σ v k :
      (V v, f_appR (e \ σ) :: k) → (C e ∣ v :: σ, k)
  | ss_pairL e₁ e₂ σ k :
      (C e₁ ∥ e₂ ∣ σ, k) → (C e₁ ∣ σ, f_pairL e₂ σ :: k)
  | ss_pairR v e σ k :
      (V v, f_pairL e σ :: k) → (C e ∣ σ, f_pairR v :: k)
  | ss_pair v₁ v₁' k :
      (V v₁', f_pairR v₁ :: k) → (V ⟨v₁, v₁'⟩, k)
  | ss_fstK e σ k :
      (C π₁ e ∣ σ, k) → (C e ∣ σ, f_fst :: k)
  | ss_fst v₁ v₁' k :
      (V ⟨v₁, v₁'⟩, f_fst :: k) → (h_val v₁, k)
  | ss_sndK e σ k :
      (C π₂ e ∣ σ, k) → (C e ∣ σ, f_snd :: k)
  | ss_snd v₁ v₁' k :
      (V ⟨v₁, v₁'⟩, f_snd :: k) → (V v₁', k)
  where "c₁ → c₂" := (sstep c₁ c₂).

  Reserved Notation "c '⇒s' v ; g" (at level 70, no associativity).
  Inductive seqbig : conf -> val -> spdag -> Prop :=
  | seq_val v :
      (V v, nil) ⇒s v; 0%sp
  | seq_step c₁ c₂ v g
             (HR : c₁ → c₂)
             (HB : c₂ ⇒s v; g) :
      c₁ ⇒s v; (1 · g)%sp
  where "c '⇒s' v ; g" := (seqbig c v g).

  Reserved Notation "c '⇒p' v ; g" (at level 70, no associativity).
  Inductive parbig : conf -> val -> spdag -> Prop :=
  | par_val v :
      (V v, nil) ⇒p v; 0%sp
  | par_step c₁ c₂ v g
             (HNP : isPar c₁ = false)
             (HR : c₁ → c₂)
             (HB : c₂ ⇒p v; g) :
      c₁ ⇒p v; (1 · g)%sp
  | par_par e₁ e₂ σ k v₁ v₁' v g₁ g₂ g
            (HB1 : (C e₁ ∣ σ, nil) ⇒p v₁; g₁)
            (HB2 : (C e₂ ∣ σ, nil) ⇒p v₁'; g₂)
            (HB  : (V ⟨v₁, v₁'⟩, k) ⇒p v; g) :
      (C e₁ ∥ e₂ ∣ σ, k) ⇒p v; ((g₁ ∥ g₂) · g)%sp
  where "c '⇒p' v ; g" := (parbig c v g).

  Reserved Notation " c @ n , N '⇒a' v ; g" (at level 70, no associativity).
  Inductive amobig (N : nat) : conf -> nat -> val -> spdag -> Prop :=
  | amo_val n v :
      (V v, nil) @ n, N ⇒a v; 0%sp
  | amo_step n h k c v g
             (Hcan : n < N \/ hasPar k = false)
             (HR : (h, k) → c)
             (HB : c @ S n, N ⇒a v; g) :
      (h, k) @ n, N ⇒a v; (1 · g)%sp
  | amo_prom n h k₁ k₂ e σ v₁ v₁' v g₁ g₂ g
             (HF : N <= n) (HS : hasPar k₂ = false)
             (HB1 : (h, k₁) @ 0, N ⇒a v₁; g₁)
             (HB2 : (h_clo e σ, nil) @ 0, N ⇒a v₁'; g₂)
             (HB : (V ⟨v₁, v₁'⟩, k₂) @ 0, N ⇒a v; g) :
      (h, k₁ ++ f_pairL e σ :: k₂) @ n, N ⇒a v; ((g₁ ∥ g₂) · g)%sp
  where "c @ n , N '⇒a' v ; g" := (amobig N c n v g).

End Semantics.

Arguments sstep (c₁ c₂)%hd.
Notation "c₁ → c₂" := (sstep c₁ c₂) (at level 70, no associativity).

Arguments seqbig (c)%hd (v)%val (g)%sp.
Notation "c '⇒s' v ; g" := (seqbig c v g) (at level 70, no associativity).

Arguments parbig (c)%hd (v)%val (g)%sp.
Notation "c '⇒p' v ; g" := (parbig c v g) (at level 70, no associativity).

Arguments amobig N (c)%hd n (v)%val (g)%sp.
Notation "c @ n , N '⇒a' v ; g" := (amobig N c n v g) (at level 70, no associativity).

Local Open Scope hd_scope.

Require Import Arith.
Require Import Omega.
Require Import Setoid.

Definition work_bound wa ws N n:=
  N * wa <= (N + τ) * ws + n * τ.

Lemma sstep_cont h h' k₁ k₁' k₂
      (HS : (h, k₁) → (h', k₁')) :
  (h, k₁ ++ k₂) → (h', k₁' ++ k₂).
Proof.
  inversion HS; subst; clear HS; eauto using sstep.
Qed.

Open Scope hd_scope.
Open Scope val_scope.
Open Scope exp_scope.

Lemma seq_cont h k₁ k₂ v₁ v₂ g₁ g₂
      (HB  : (h, k₁) ⇒s v₁; g₁)
      (HBC : (V v₁, k₂) ⇒s v₂; g₂) :
  exists g,
    (h, k₁ ++ k₂) ⇒s v₂; g /\
    work g = work g₁ + work g₂.
Proof.
  remember (h, k₁) as c; revert h k₁ Heqc.
  induction HB; intros; subst; [inversion_clear Heqc; simpl; now eauto |].
  destruct c₂ as [h' k']; specialize (IHHB HBC _ _ (eq_refl _)); destruct IHHB as [g' [HB' HW]].
  eexists; split; [now eauto using seqbig, sstep_cont | simpl].
  congruence.
Qed.

Lemma seq_cut h k₁ k₂ e σ v₁ v₂ v g₁ g₂ g
      (HB1 : (h, k₁) ⇒s v₁; g₁)
      (HB2 : (h_clo e σ, nil) ⇒s v₂; g₂)
      (HB3 : (V ⟨v₁, v₂⟩, k₂) ⇒s v; g) :
  exists g',
    (h, k₁ ++ f_pairL e σ :: k₂) ⇒s v; g' /\
    work g' = work g₁ + work g₂ + work g + 2.
Proof.
  eapply seq_step in HB3; [| now apply ss_pair].
  eapply seq_cont in HB2; [| exact HB3].
  destruct HB2 as [g23 [HB23 HW23]]; clear HB3.
  eapply seq_step in HB23; [| now apply ss_pairR].
  eapply seq_cont in HB1; [| exact HB23].
  clear HB23; destruct HB1 as [g' [HB HW]]; eexists; split; [exact HB |].
  simpl in *; rewrite HW, HW23; clear HW HW23.
  omega.
Qed.

Lemma work_amo N c n v ga
      (HB : c @ n, N ⇒a v; ga) :
  exists gs, c ⇒s v; gs /\ work_bound (work ga) (work gs) N n.
Proof.
  induction HB.
  - exists 0%sp; split; [now eauto using seqbig | unfold work_bound; simpl].
    rewrite !mult_0_r; eauto with arith.
  - destruct IHHB as [gs [HBS HWB]].
    exists (1 · gs)%sp; split; [now eauto using seqbig | unfold work_bound in *; simpl].
    rewrite <- !mult_n_Sm; simpl in HWB; omega.
  - destruct IHHB1 as [gs₁ [HBS1 HWB1]]; destruct IHHB2 as [gs₂ [HBS2 HWB2]];
    destruct IHHB3 as [gs [HBS HWB]].
    lets (g' & HB & HW): seq_cut HBS1 HBS2 HBS.
    eexists; split; [exact HB |].
    rewrite HW; clear -HWB1 HWB2 HWB HF; unfold work_bound in *; simpl in *.
    rewrite plus_0_r in *.
    rewrite <- !mult_n_Sm, !Nat.mul_add_distr_l.
    lets: mult_le_compat_r τ HF.
    omega.
Qed.

Theorem work_amortized N e v ga
        (HB : (C e ∣ nil, nil) @ 0, N ⇒a v; ga) :
  exists gs,
    (C e ∣ nil, nil) ⇒s v; gs /\
    N * (work ga) <= (N + τ) * (work gs).
Proof.
  apply work_amo in HB; destruct HB as [gs [HB HW]]; exists gs; split; [assumption |].
  unfold work_bound in HW; simpl in HW; rewrite plus_0_r in HW; assumption.
Qed.

Reserved Notation "c ⇒p' v ; g" (at level 70, no associativity).
Inductive parbig' : conf -> val -> spdag -> Prop :=
| par'_simp h k v g
            (HK : hasPar k = false)
            (HB : (h, k) ⇒p v; g) :
    (h, k) ⇒p' v; g
| par'_par h k₁ k₂ e σ v₁ v₂ v g₁ g₂ g (HK : hasPar k₂ = false)
           (HB1 : (h, k₁) ⇒p' v₁; g₁)
           (HB2 : (C e ∣ σ, nil) ⇒p v₂; g₂)
           (HB  : (V ⟨v₁, v₂⟩, k₂) ⇒p v; g) :
    (h, k₁ ++ f_pairL e σ :: k₂) ⇒p' v; ((g₁ ∥ g₂) · g)%sp
where "c ⇒p' v ; g" := (parbig' c v g).

Definition span_bound sa sp N n (hasPars : bool) :=
  τ * (sa + if hasPars then min N n else 0) <= (N + τ) * sp.


Lemma parbig_seq_split h k₁ k₂ v g
      (HK : hasPar k₂ = false)
      (HB : (h, k₁ ++ k₂) ⇒p v; g) :
  exists v' g₁ g₂, (h, k₁) ⇒p v'; g₁ /\ (V v', k₂) ⇒p v; g₂
                   /\ span g₁ + span g₂ = span g.
Proof.
  remember (h, k₁ ++ k₂) as c; revert h k₁ Heqc.
  induction HB; intros; inversion Heqc; clear Heqc; subst.
  - lets (H2 & H3): nil_eq_app_inv H1; subst.
    repeat eexists; eauto using parbig.
  - inversion HR; subst h c₂; clear HR; simpl in *; try discriminate;
    match type of H1 with
    (* rules that inspect the context: in case k₁ is empty, these work by using
       the value reduction in the left parbig and replaying the step on the right,
       otherwise by induction, like all the rules that simply extend the context
       or do nothing to it *)
    | ?f :: ?k = k₁ ++ k₂ =>
      destruct k₁ as [| f₁ k₁]; rew_list in *; [subst k₂ | inversion H1; clear H1; subst f₁ k]
    | ?k = k₁ ++ k₂ => subst k
    end;
    try (specialize (IHHB _ _ (eq_refl _)) || specialize (IHHB _ (_ :: _) (eq_refl _));
         destruct IHHB as [v' [g₁ [g₂ [HB1 [HB2 HW]]]]]);
    (repeat eexists; now eauto using parbig, sstep || simpl; congruence).
  - clear IHHB1 IHHB2 HB3; specialize (IHHB3 _ _ (eq_refl _)).
    destruct IHHB3  as [v' [g₃ [g₄ [HB3 [HB4 HW]]]]].
    exists v'. exists ((g₁ ∥ g₂) · g₃)%sp. exists g₄.
    repeat split; [now eauto using parbig | assumption | simpl].
    now rewrite <- HW, plus_assoc.
Qed.

Lemma parbig'_seq_parbig h k v g (HS : hasPar k = false)
      (HB : (h, k) ⇒p' v; g) :
  (h, k) ⇒p v; g.
Proof.
  inversion HB; subst; clear HB; [assumption |].
  rewrite hasPar_app_par in HS; discriminate.
Qed.

Lemma parbig'_par e₁ e₂ σ k v g
      (HB : (C e₁ ∣ σ, f_pairL e₂ σ :: k) ⇒p' v; g) :
  (C e₁ ∥ e₂ ∣ σ, k) ⇒p' v; g.
Proof.
  remember (h_clo e₁ σ, f_pairL e₂ σ :: k) as c; revert k Heqc.
  induction HB; intros; [inversion Heqc; subst; discriminate |].
  destruct k₁ as [| f k₁]; rew_list in Heqc; inversion Heqc; clear Heqc; subst; [| now eauto using parbig'].
  apply parbig'_seq_parbig in HB; [| reflexivity].
  eauto using parbig', parbig.
Qed.

Lemma span_bound_par_small sa sp N n (HN : n < N)
      (HS : span_bound sa sp N (S n) true) :
  span_bound (S sa) sp N n true.
Proof.
  unfold span_bound in *; simpl in *; unfold lt in HN.
  rewrite <- HS; clear HS.
  rewrite !Min.min_r by auto with arith.
  auto with arith.
Qed.

Lemma parbig'_par_step_bound h k c v g
      (HNP : isPar (h, k) = false)
      (HS  : (h, k) → c)
      (HB  : c ⇒p' v; g) :
  exists g', (h, k) ⇒p' v; g' /\ span g <= span g'.
Proof.
  revert h k HNP HS; induction HB; intros; [remember (isParL (h0, k0)) as b; destruct b |].
  (* isParL = true *)
  - inversion HS; subst; try discriminate; []; clear HS Heqb.
    change (f_pairL e σ :: k1) with (nil ++ f_pairL e σ :: k1).
    apply parbig_seq_split with (k₁ := nil) in HB; [| assumption].
    rename v0 into v₁; destruct HB as [v₂ [g₂ [g₃ [HB2 [HBC HS]]]]].
    inversion HBC; inversion HR; clear HNP HBC HR; subst.
    eexists; split; [now eauto using parbig', parbig |].
    rewrite <- HS; clear; simpl; omega.
  - assert (HT : hasPar k0 = false) by
        (inversion HS; subst; clear HS; simpl in *; assumption || discriminate).
    eexists; split; [now eauto using parbig', parbig |].
    simpl; now auto with arith.
  - assert (HT : exists k', sstep (h0, k') (h, k₁) /\ k = k' ++ f_pairL e σ :: k₂).
    { clear -HNP HS.
      inversion HS; subst h0; simpl in *; try discriminate;
      match type of H3 with
      | ?f :: ?k = k₁ ++ ?k' => destruct k₁ as [| f' k₁]; inversion H3
      | _ => idtac
      end; subst;
      (eexists; split; [now eauto using sstep | reflexivity]).
    }
    destruct HT as [k' [HS' HK']]; subst k.
    edestruct IHHB as [g' [HN' HSP]]; [simpl in *; exact HNP | eassumption |].
    eexists; split; [now eauto using parbig' |].
    simpl; now apply le_n_S, plus_le_compat_r, plus_le_compat_l, Nat.max_le_compat_r.
Qed.

Lemma span_amo N h k n v ga
      (HN : 1 <= N)
      (HB : (h, k) @ n, N ⇒a v; ga) :
  exists gp, (h, k) ⇒p' v; gp /\ span_bound (span ga) (span gp) N n (hasPar k).
Proof.
  remember (h, k) as c; revert h k Heqc.
  induction HB; intros; inversion Heqc; subst; clear Heqc || clear H.
  - eexists; split; [now eauto using parbig', parbig |].
    unfold span_bound; simpl; rewrite !mult_0_r; reflexivity.
  - destruct c as [h' k']; specialize (IHHB _ _ (eq_refl _)); destruct IHHB as [gp [HBS HW]].
    destruct (isPar (h0, k0)) eqn: EQisPar.
    + inversion HR; subst; simpl in EQisPar; try discriminate; []; clear HR EQisPar; simpl in HW.
      exists gp; split; [now apply parbig'_par | clear - HW Hcan HN; simpl].
      destruct (hasPar k0); [destruct Hcan; [| discriminate]; now apply span_bound_par_small |].
      unfold span_bound in *; rewrite <- HW, plus_0_r; clear HW Hcan.
      rewrite <- Nat.add_1_r; apply mult_le_compat_l, plus_le_compat_l.
      apply Min.min_glb; now auto with arith.
    + assert (HT : hasPar k' = hasPar k0 \/ isParL (h0, k0) = true).
      { inversion HR; subst; clear HR; simpl in *; discriminate || tauto. }
      destruct (hasPar k') eqn: EQCS; [| destruct HT as [HT | HT]; [symmetry in HT |]].
      * eapply parbig'_par_step_bound in HBS; [| eassumption | eassumption].
        destruct HBS as [g' [HBS HS]]; eexists; split; [eassumption |].
        assert (EQCS' : hasPar k0 = true) by
            (destruct HT as [HT | HT]; [congruence | inversion HR; subst; simpl in HT; discriminate || reflexivity]).
        rewrite EQCS' in *; destruct Hcan as [HNN | HD]; [clear -HS HW HNN HN | discriminate].
        unfold span_bound in *; simpl in *.
        etransitivity; [| apply mult_le_compat_l, HS]; rewrite <- HW; clear HW HS.
        rewrite plus_n_Sm; apply mult_le_compat_l, plus_le_compat_l.
        rewrite !Nat.min_r by auto with arith; reflexivity.
      * rewrite HT in *; apply parbig'_seq_parbig in HBS; [| assumption].
        eexists; split; [now eauto using parbig', parbig |].
        clear -HW HN; unfold span_bound in *; simpl in *; rewrite plus_0_r in *.
        rewrite <- !mult_n_Sm; omega.
      * inversion HR; subst; simpl in *; try discriminate; [].
        destruct Hcan as [HNN | HD]; [| discriminate].
        clear -HN HNN HW HBS EQCS; apply parbig'_seq_parbig in HBS; [| assumption].
        apply parbig_seq_split with (k₁ := nil) in HBS; [| assumption].
        destruct HBS as [v₂ [g₂ [g₃ [HB2 [HBC HS]]]]].
        inversion HBC; subst; inversion HR; subst; clear HR HBC.
        change (f_pairL e σ :: k) with (nil ++ f_pairL e σ :: k).
        eexists; split; [now eauto using parbig', parbig |].
        clear -HN HNN HW HS; unfold span_bound in *; simpl in *.
        rewrite <- HS, plus_0_r in HW; clear HS.
        rewrite !plus_n_Sm, <- plus_assoc, Nat.mul_add_distr_l, Nat.mul_add_distr_l.
        etransitivity; [| apply plus_le_compat_l, HW].
        rewrite plus_comm, mult_comm with (m := τ), Nat.min_r by auto with arith.
        apply plus_le_compat_r, mult_le_compat_l.
        transitivity N; now auto with arith.
  - forwards~ (gp1 & HBS1 & HW1): (rm IHHB1).
    forwards~ (gp2 & HBS2 & HW2): (rm IHHB2).
    forwards~ (gp3 & HBS3 & HW3): (rm IHHB3).

    inverts HBS2; [| destruct k₁0; simpl in H1; discriminate].
    inverts HBS3; [| rewrite hasPar_app_par in HS; discriminate].
    eexists; split; [now eauto using parbig' |].
    clear -HF HW1 HW2 HW3 HK0.
    unfold span_bound in *; simpl in *; rewrite hasPar_app_par; rewrite HK0 in HW3; clear HK0.
    rewrite <- !mult_n_Sm, !Nat.mul_add_distr_l.
    rewrite Min.min_0_r in HW1.
    assert (HK : (if hasPar k₁ then 0 else 0) = 0) by now destruct (hasPar k₁).
    rewrite HK in HW1; clear HK.
    rewrite !plus_0_r in *.
    rewrite <- !Nat.mul_max_distr_l, Min.min_l by assumption.
    rewrite Nat.mul_add_distr_r with (p := τ), (mult_comm τ N).
    etransitivity.
    + apply plus_le_compat_r, plus_le_compat_r.
      apply plus_le_compat; [| eassumption].
      apply plus_le_compat_l, Nat.max_le_compat; eassumption.
    + omega.
Qed.

Theorem span_amortized N e v ga
        (HN : 1 <= N)
        (HB : (C e ∣ nil, nil) @ 0, N ⇒a v; ga) :
  exists gs,
    (C e ∣ nil, nil) ⇒p v; gs /\
    τ * (span ga) <= (N + τ) * (span gs).
Proof.
  apply span_amo in HB; [| assumption].
  destruct HB as [gs [HB HW]]; inversion HB; subst; clear HB.
  - eexists; split; [eassumption |].
    clear -HW; unfold span_bound in HW; simpl in HW.
    now rewrite plus_0_r in HW.
  - destruct k₁; simpl in HK; discriminate.
Qed.

(** SOUNDNESS PART **)

Lemma existsPar :
  forall c,
    (isPar c = false)
    \/
    (exists e₁ e₂ σ k, c = ((C e₁ ∥ e₂ ∣ σ), k)).
Proof.
  destruct c as [h k].
  destruct h; (try destruct e); try (left; reflexivity).
  right; eauto.
Qed.

Lemma sstep_case h k h' k'
      (HS : (h, k) → (h', k')) :
  (exists v, h' = V v /\ k' = k)
  \/
  (exists f, k' = f :: k)
  \/
  (exists f v, h = V v /\ k = f :: k')
  \/
  (exists f f' v k'', h = V v /\ k = f :: k'' /\ k' = f' :: k'').
Proof.
  inverts keep HS; eauto 10.
Qed.

Ltac list_refl := rew_list; reflexivity.

Lemma app_eq_suff_one_inv {A} (l₁ l₁' : list A) a :
      l₁ ++ (a :: nil) = l₁' ++ (a :: nil) -> l₁ = l₁'.
Proof.
  gen l₁'; induction l₁; introv; intro Heq; rew_list in *.
  - destruct l₁'; auto.
    rew_list in Heq.
    inverts Heq; false; eapply nil_eq_last_inv; eauto.
  - destruct l₁'; rew_list in *.
    + inverts Heq; false; eapply nil_eq_last_inv; eauto.
    + inverts Heq; f_equal.
      eauto.
Qed.

Lemma app_eq_suff_inv {A} (l₁ l₁' l₂ : list A) :
      l₁ ++ l₂ = l₁' ++ l₂ -> l₁ = l₁'.
Proof.
  gen l₁ l₁'; induction l₂; introv; intro Heq.
  - rew_list in Heq; assumption.
  - forwards: IHl₂ (l₁ ++ (a :: nil)) (l₁' ++ (a :: nil)).
    + rew_list; assumption.
    + eapply app_eq_suff_one_inv.
      eauto.
Qed.

Ltac suff_inv H :=
  match type of H with
    | ?f :: ?k₁ ++ ?k₂ = ?k₁' ++ ?k₂ =>
      replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in H by list_refl;
      apply app_eq_suff_inv in H
    | ?k₁ ++ ?k₂ = ?k₁' ++ ?k₂ =>
      eapply app_eq_suff_inv in H
    | ?k = ?k' ++ ?f :: ?k =>
      replace (k' ++ f :: k) with ((k' & f) ++ k) in H by list_refl;
      replace k with (nil ++ k) in H at 1 by list_refl;
      symmetry in H; apply app_eq_suff_inv in H
    | _ =>
      fail "not invertible"
  end.

Ltac suff_inv_all :=
  repeat match goal with | [ H : _ |- _ ] => suff_inv H end.

Ltac uncons_lists :=
  repeat
    match goal with
    | [ H : ?f₁ :: ?l₁ = ?f₂ :: ?l₂ |- _ ] =>
      inverts H
    end.

Lemma sstep_split h h' k₁ k₁' k₂
      (HS : (h, k₁ ++ k₂) → (h', k₁' ++ k₂)) :
  (h, k₁) → (h', k₁').
Proof.
  inverts HS as; [intro Hnth | ..]; intros;
    suff_inv_all; subst; eauto using sstep; [|];
      (destruct k₁; destruct k₁'; rew_list in *; try subst; try discriminate;
       uncons_lists; suff_inv_all; subst;
       eauto using sstep || false; eauto using last_eq_nil_inv).
Qed.

Lemma sstep_split_nil h h' k
      (HS : (h, k) → (h', k)) :
  (h, nil) → (h', nil).
Proof.
  eapply sstep_split; eauto.
Qed.

Lemma notPar_app h k₁ k₂
      (HP : isPar (h, k₁) = false) :
  isPar (h, k₂) = false.
Proof.
  destruct h as [e | ?]; [destruct e | ]; simpl in *; congruence.
Qed.

Lemma parbig_split h k₁ k₂ v₂ g
      (HP : (h, k₁ ++ k₂) ⇒p v₂; g) :
  exists v₁ g₁ g₂, (h, k₁) ⇒p v₁; g₁ /\ (V v₁, k₂) ⇒p v₂; g₂.
Proof.
  remember (h, k₁ ++ k₂) as c₁; gen h k₁ k₂; induction HP; introv; intro Heq.
  - inverts Heq as Heq; destruct (app_eq_nil_inv (eq_sym Heq)); subst; repeat constructors.
  - destruct c₁; inverts Heq. destruct c₂ as [h' k'].
    forwards [(v' & ? & ?) | [(f & ?) | [(f & v' & ? & Hk) | (f & f' & v' & k'' & ? & Hk & ?)]]]: sstep_case HR; subst.
    + forwards~ (v₁ & g₁ & g₂ & HP1 & HP2): IHHP (V v'); clear IHHP.
      apply sstep_split in HR.
      forwards: par_step HR HP1; try eapply notPar_app; eauto.
    + forwards~ (v₁ & g₁ & g₂ & HP1 & HP2): IHHP h' (f :: k₁) k₂; clear IHHP.
      replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in HR by list_refl.
      apply sstep_split in HR.
      forwards: par_step HR HP1; try eapply notPar_app; eauto.
    + destruct k₁; rew_list in *; subst.
      * forwards* (v₁ & g₁ & g₂ & HP1 & HP2): IHHP h' k' (@nil frame); [list_refl | ]; clear IHHP.
        {
          replace v₁ with v in *; [clear HP2|].
          - forwards*: par_step HR HP1.
            eauto using parbig.
          - inverts* HP2 as.
            intros ? ? Habs.
            inverts Habs.
        }
      * replace f0 with f in * by congruence.
        injection Hk; clear Hk; intro Heq; subst.
        forwards* (v₁ & g₁ & g₂ & HP1 & HP2): IHHP k₁ k₂; clear IHHP.
        replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in HR by list_refl.
        apply sstep_split in HR.
        forwards*: par_step HR HP1.
    + destruct k₁; rew_list in *; subst.
      * forwards* (v₁ & g₁ & g₂ & HP1 & HP2): IHHP h' (f' :: k'') (@nil frame); [list_refl | ]; clear IHHP.
        {
          replace v₁ with v in *; [clear HP2|].
          - forwards*: par_step HR HP1.
            eauto using parbig.
          - inverts* HP2 as.
            intros ? ? Habs.
            inverts Habs.
        }
      * replace f0 with f in * by congruence.
        injection Hk; clear Hk; intro Heq; subst.
        forwards~ (v₁ & g₁ & g₂ & HP1 & HP2): IHHP (f' :: k₁) k₂; clear IHHP.
        replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in HR by list_refl.
        replace (f' :: k₁ ++ k₂) with ((f' :: k₁) ++ k₂) in HR by list_refl.
        apply sstep_split in HR.
        forwards*: par_step HR HP1.
  - forwards~ (va & ga1 & ga2 & HPa1 & HPa2): IHHP1 (@nil frame) (@nil frame); clear IHHP1.
    forwards~ (vb & gb1 & gb2 & HPb1 & HPb2): IHHP2 (@nil frame) (@nil frame); clear IHHP2.
    inverts HPa2 as ? Habs; [|now invert Habs].
    inverts HPb2 as ? Habs; [|now invert Habs].
    inverts Heq.
    forwards~ (vc & gc1 & gc2 & HPc1 & HPc2): IHHP3 k₁ k₂; clear IHHP3.
    forwards~: par_par HPa1 HPb1 HPc1.
    eauto.
Qed.

Lemma parbig_par e₁ e₂ σ k v g
      (HP : (C e₁ ∣ σ, f_pairL e₂ σ :: k) ⇒p v; g) :
  exists g', (C (e₁ ∥ e₂) ∣ σ, k) ⇒p v; g'.
Proof.
  replace (f_pairL e₂ σ :: k) with ((f_pairL e₂ σ :: nil) ++ k) in HP by list_refl.
  forwards (v₁₂ & ? & gk & HP1' & HPk): parbig_split HP; clear HP.
  replace (f_pairL e₂ σ :: nil) with (nil ++ f_pairL e₂ σ :: nil) in HP1' by list_refl.
  forwards (v₁ & g₁ & g₂ & HP1 & HP2): parbig_split HP1'; clear HP1'.
  inverts HP2 as [HR Hpair].
  destruct c₂ as [h₂ k₂].
  inverts HR.
  replace (f_pairR v₁ :: nil) with (nil ++ f_pairR v₁ :: nil) in Hpair by list_refl.
  forwards (v₂ & ? & ? & HP2 & Hpair'): parbig_split Hpair; clear Hpair.
  inverts Hpair' as [HR].
  inverts HR.
  inverts HB.
  - eauto using parbig.
  - inverts HR.
Qed.

Theorem seq_par c v g
        (HS : c ⇒s v; g) :
  exists g', c ⇒p v; g'.
Proof.
  induction HS.
  - eexists; constructor.
  - destruct IHHS as [g' HP].
    forwards [HnoPar | (e₁ & e₂ & σ & k₁ & ?)]: existsPar c₁.
    + eexists; eapply par_step; eauto.
    + subst.
      inverts HR.
      apply parbig_par in HP.
      assumption.
Qed.

Lemma sstep_no_pair_no_par h k h' k'
      (HS : (h, k) → (h', k'))
      (Hnoparh : isPar (h, k) = false)
      (Hnopark: hasPar k = false) :
  hasPar k' = false.
Proof.
  inverts HS; auto || discriminate.
Qed.

Lemma pair_amo_nopar h k v g
      (HP : (h, k) ⇒p v; g)
      (Hnopar : hasPar k = false) :
  forall n, exists g', (h, k) @ n, 0 ⇒a v; g'.
Proof.
  remember (h, k) as c eqn:Heq; gen h k; induction HP; introv; intros;
  try (injection Heq; clear Heq); intros; subst.
  - eauto using amobig.
  - destruct c₂ as (h', k').
    forwards H: sstep_no_pair_no_par HR HNP Hnopar.
    forwards~ (g' & HP'): IHHP h' k' H (S n).
    eauto using amobig.
  - idtac.
    forwards~ (g₁' & HP1'): IHHP1 0.
    forwards~ (g₂' & HP2'): IHHP2 0.
    forwards~ (gk & Hk): IHHP3 0.
    eexists; eapply amo_step; [now auto | now constructor | ].
    replace (f_pairL e₂ σ :: k0) with (nil ++ f_pairL e₂ σ :: k0) by list_refl.
    eauto 2 using amobig with arith.
Qed.

Lemma sstep_pairL_inv_right h k h' k₁ e₂ σ k₂
      (HR : (h, k) → (h', k₁ ++ f_pairL e₂ σ :: k₂)) :
  (exists v, h' = V v /\ k = k₁ ++ f_pairL e₂ σ :: k₂)
  \/
  (exists k₁', k = k₁' ++ f_pairL e₂ σ :: k₂)
  \/
  (k₁ = nil /\ k = k₂).
Proof.
  inverts keep HR as; eauto; try intro Heq;
  destruct k₁ as [|f k₁']; try (rew_list in Heq; inverts Heq); eauto;
  try
    (
      branch 2;
      match goal with
        | [ |- exists k₁'0, ?f :: ?k ++ f_pairL e₂ σ :: k₂ = k₁'0 ++ f_pairL e₂ σ :: k₂ ] =>
          replace (f :: k ++ f_pairL e₂ σ :: k₂) with ((f :: k) ++ f_pairL e₂ σ :: k₂) by list_refl
      end;
      eauto
    ).
Qed.

Lemma sstep_pairL_inv_left h k₁ e₂ σ k₂ h' k'
      (HR : (h, k₁ ++ f_pairL e₂ σ :: k₂) → (h', k')) :
  (exists v, h' = V v /\ k' = k₁ ++ f_pairL e₂ σ :: k₂)
  \/
  (exists k₁', k' = k₁' ++ f_pairL e₂ σ :: k₂)
  \/
  (exists v₁, h = V v₁ /\ k₁ = nil /\ k' = f_pairR v₁ :: k₂).
Proof.
  inverts keep HR as; eauto; try intro Heq;
  destruct k₁ as [|f k₁']; try (rew_list in Heq; inverts Heq); eauto;
  (
    branch 2;
    match goal with
      | [ |- exists k₁'0, ?f :: ?k ++ f_pairL e₂ σ :: k₂ = k₁'0 ++ f_pairL e₂ σ :: k₂ ] =>
        replace (f :: k ++ f_pairL e₂ σ :: k₂) with ((f :: k) ++ f_pairL e₂ σ :: k₂) by list_refl
    end;
    eauto
  )
    ||
  (
    branch 3;
    now eauto
  ).
Qed.

Lemma to_have_or_not_to_have_par k :
  (hasPar k = false)
  \/
  (exists k₁ e₂ σ k₂, k = k₁ ++ f_pairL e₂ σ :: k₂ /\ hasPar k₂ = false).
Proof.
  induction k.
  - auto.
  - destruct IHk as [Hnopark | (k₁ & e₂ & σ & k₂ & ? & Hnopark2)]; subst.
    + destruct a; simpl in *; auto.
      right; exists (@nil frame) e s k; auto.
    + right; exists (a :: k₁) e₂ σ k₂; auto.
Qed.

Lemma sstep_split_case h k₁ k₂ h' k'
      (HR : (h, k₁ ++ k₂) → (h', k')) :
  (exists k₁', k' = k₁' ++ k₂) \/
  (exists f, k₁ = nil /\ k' = f :: k₂) \/
  (exists v f k₂', k₁ = nil /\ k₂ = f :: k₂' /\ k' = k₂' /\ h = V v) \/
  (exists v f f' k₂', k₁ = nil /\ k₂ = f :: k₂' /\ k' = f' :: k₂' /\ h = V v).
Proof.
  forwards [(v & ? & ?) | [(f & ?) | [(f & v & ? & Heq) | (f & f' & v & k'' & ? & Heq & ?)]]] : sstep_case HR; subst.
  - eauto.
  - branch 1; eexists (_ :: _); reflexivity.
  - destruct k₁; rew_list in *; subst; [|inverts Heq]; [branch 3|]; eauto 10.
  - destruct k₁; rew_list in *; subst; [|inverts Heq]; [branch 4|]; eauto 10.
    branch 1; eexists (_ :: _); reflexivity.
Qed.

Ltac normalize_ssteps :=
  repeat
    match goal with
      | [ H : (_, ?f :: ?k₁ ++ ?k₂) → (_, _) |- _ ] =>
        replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in H by list_refl
      | [ H : (_, _) → (_, ?f :: ?k₁ ++ ?k₂) |- _ ] =>
        replace (f :: k₁ ++ k₂) with ((f :: k₁) ++ k₂) in H by list_refl
      | [ H : (_, ?k) → (_, ?f :: ?k) |- _ ] =>
        replace k with (nil ++ k) in H at 1 by list_refl;
        replace (f :: k) with ((f :: nil) ++ k) in H by list_refl
      | [ H : (_, ?f :: ?k) → (_, ?k) |- _ ] =>
        replace k with (nil ++ k) in H at 2 by list_refl;
        replace (f :: k) with ((f :: nil) ++ k) in H by list_refl
    end.

Lemma seq_split h k₁ k₂ v g
      (HS : (h, k₁ ++ k₂) ⇒s v; g) :
  exists v' g₁ g₂, (h, k₁) ⇒s v'; g₁ /\ (V v', k₂) ⇒s v; g₂ /\ work g = work g₁ + work g₂.
Proof.
  remember (h, k₁ ++ k₂) as c; gen h k₁ k₂; induction HS; introv; intros Heq.
  - inverts Heq as; intro Heq; destruct (nil_eq_app_inv Heq); subst.
    eauto 10 using seqbig.
  - subst; destruct c₂ as [h' k'].
    forwards [(k₁' & ?) | [(f & ? & ?) | [(v' & f & k₂' & ? & ? & ? & ?) | (v' & f & f' & k₂' & ? & ? & ? & ?)]]] : sstep_split_case HR; rew_list in *; subst; rew_list in *.
    + forwards~ (v' & g₁ & g₂ & HS1' & HS2 & ?): IHHS k₁' k₂.
      apply sstep_split in HR.
      forwards: seq_step HR HS1'.
      exists v' (1 · g₁)%sp g₂.
      simpl; intuition omega.
    + forwards~ (v' & g₁ & g₂ & HS1' & HS2 & ?): IHHS (f :: nil) k₂.
      normalize_ssteps; apply sstep_split in HR; forwards: seq_step HR HS1'.
      exists v' (1 · g₁)%sp g₂.
      simpl; intuition omega.
    + forwards~ (v'' & g₁ & g₂ & HS1' & HS2 & ?): IHHS (@nil frame) k₂'.
      forwards (g' & HS' & ?): seq_cont HS1' HS2; rew_list in *.
      forwards: seq_step HR HS'.
      exists v' 0%sp (1 · g')%sp.
      simpl; repeat split; eauto using seqbig || omega.
    + forwards~ (v'' & g₁ & g₂ & HS1' & HS2 & ?): IHHS (@nil frame) (f' :: k₂').
      forwards (g' & HS' & ?): seq_cont HS1' HS2; rew_list in *.
      forwards: seq_step HR HS'.
      exists v' 0%sp (1 · g')%sp.
      simpl; repeat split; eauto using seqbig || omega.
Qed.

Lemma seq_split_pairl h k₁ e₂ σ k₂ v g
      (HS : (h, k₁ ++ f_pairL e₂ σ :: k₂) ⇒s v; g) :
  exists v₁ g₁ v₂ g₂ g₃,
    ((h, k₁) ⇒s v₁; g₁) /\ (work g₁ < work g) /\
    ((C e₂ ∣ σ, nil) ⇒s v₂; g₂) /\ (work g₂ < work g) /\
    ((V ⟨v₁, v₂⟩, k₂) ⇒s v; g₃) /\ (work g₃ < work g).
Proof.
  forwards (v₁ & g₁ & g' & HS1 & HS' & ?): seq_split HS.
  inverts HS' as.
  intros [h' k'] HR HS''.
  inverts HR as.
  replace (f_pairR v₁ :: k₂) with (nil ++ f_pairR v₁ :: k₂) in HS'' by list_refl.
  forwards (v₂ & g₂ & g₃ & HS2 & HS''' & ?): seq_split HS''.
  inverts HS''' as.
  intros ? HR HS3.
  inverts HR.
  exists v₁ g₁ v₂ g₂ g1.
  simpl in *; intuition omega.
Qed.

Require Import Coq.Init.Wf.

Lemma prom_or_not_prom k n T :
  (exists k₁ e₂ σ k₂, k = k₁ ++ f_pairL e₂ σ :: k₂ /\ hasPar k₂ = false /\ T <= n) \/
  (n < T \/ hasPar k = false).
Proof.
  forwards [Hnopar | (k₁ & e₂ & σ & k₂ & ? & ?)]: to_have_or_not_to_have_par k; subst;
  assert (n < T \/ T <= n) as [|] by omega;
  eauto 10.
Qed.

Theorem seq_amo c v g
      (HS : c ⇒s v; g) :
  forall n T, exists g', c @ n, T ⇒a v; g'.
Proof.
  remember (work g) as w.
  gen c v g.
  induction w as [? IH] using (well_founded_ind lt_wf); introv; intros HS Heq; introv.
  destruct c as [h k].
  forwards [(k₁ & e₂ & σ & k₂ & ? & ? & ?) | ?]: prom_or_not_prom k n T; subst.
  - forwards (v₁ & g₁ & v₂ & g₂ & g₃ & HS1 & ? & HS2 & ? & HS3 & ?): seq_split_pairl HS.
    forwards* (g₁' & HA1): IH (work g₁) HS1 0 T.
    forwards* (g₂' & HA2): IH (work g₂) HS2 0 T.
    forwards* (g₃' & HA3): IH (work g₃) HS3 0 T.
    eauto using amobig.
  - inverts HS as.
    + eauto using amobig.
    + intros [h' k'] HR HS'.
      forwards~ (g' & HA'): IH (work g0) HS' (S n) T.
      eauto using amobig.
Qed.

Theorem amo_amo T c n v g
        (HA : c @ n, T ⇒a v; g) :
  forall n' T', exists g', c @ n', T' ⇒a v; g'.
Proof.
  introv;
  forwards (? & HS & ?): work_amo HA.
  forwards: seq_amo HS n' T'.
  assumption.
Qed.

Theorem par_amo h v g
        (HP : (h, nil) ⇒p v; g) :
  forall n N, exists g', (h, nil) @ n, N ⇒a v; g'.
Proof.
  introv.
  forwards~ (g' & HA): pair_amo_nopar HP n.
  forwards~: amo_amo HA.
  eauto.
Qed.
